@@ -119,7 +119,6 @@ def process_numeric_data(
     numeric_precision: int = 4,
     transform_data: Dict = None,
 ) -> Tuple[pd.Series, Dict]:
-    series = series.round(numeric_precision).astype(str)
     is_transform = True
 
     if transform_data is None:
@@ -136,6 +135,32 @@ def process_numeric_data(
         transform_data["max_len"] = max_len
         transform_data["numeric_precision"] = numeric_precision
 
+    # Note that at this point, we should have casted int-like values to
+    # pd.Int64Dtype but just to be very sure, let's do that again here.
+    try:
+        series = series.astype(pd.Int64Dtype())
+    except TypeError:
+        pass
+    except ValueError:
+        pass
+
+    if series.dtype == pd.Int64Dtype():
+        series = series.astype(str)
+    else:
+        # We convert float-like values to string with the specified
+        # maximum precision.
+        # NOTE: Don't use series.round(numeric_precision).astype(str)
+        # In some cases, this introduces scientific notation in the
+        # string version causing "invalid" values.
+
+        # Note that our purpose for doing this is to actually truncate
+        # the precision and not increase the precision.
+        # So, we strip the right trailing zeros because the formatting
+        # pads the series to the numeric_precision even when not needed.
+        series = series.map(
+            lambda x: f"{x:.{numeric_precision}f}"
+        ).str.rstrip("0")
+
     # Get the most significant digit
     if is_transform:
         mx_sig = transform_data["mx_sig"]
@@ -265,7 +290,8 @@ def encode_column_values(series):
 
 
 def decode_column_values(series):
-    return series.str.split(SPECIAL_COL_SEP).map(lambda x: x[-1])
+    # return series.str.split(SPECIAL_COL_SEP).map(lambda x: x[-1])
+    return series.map(lambda x: x.split(SPECIAL_COL_SEP)[-1])
 
 
 def encode_processed_column(idx, dtype, col):
