@@ -44,6 +44,7 @@ class REaLSampler:
         columns: List,
         datetime_columns: List,
         column_dtypes: Dict,
+        column_has_missing: Dict,
         drop_na_cols: List,
         col_transform_data: Dict,
         random_state: Optional[int] = 1029,
@@ -61,6 +62,7 @@ class REaLSampler:
         self.datetime_columns = datetime_columns
 
         self.column_dtypes = column_dtypes
+        self.column_has_missing = column_has_missing
         self.drop_na_cols = drop_na_cols
 
         self.col_transform_data = col_transform_data
@@ -74,6 +76,9 @@ class REaLSampler:
         else:
             self.model = model
 
+        self.invalid_gen_samples = 0
+        self.total_gen_samples = 0
+
         # Set the model to eval mode
         self.model.eval()
 
@@ -112,20 +117,23 @@ class REaLSampler:
                 series = synth_df[c]
 
                 if pd.api.types.is_numeric_dtype(d):
-                    # If original data type is numeric,
-                    # attempt to replace the <NA> token
-                    # with NaN for casting to work.
-                    series = series.replace("<NA>", "NaN")
+                    if self.column_has_missing[c]:
+                        # If original data type is numeric,
+                        # attempt to replace the <NA> token
+                        # with NaN for casting to work.
+                        series = series.replace("<NA>", "NaN")
 
                     if series.dtype == "object":
                         # Explicitly unify the data type
                         # in case we have mixed types in the series.
                         series = series.astype(str)
 
-                        # Values where nan was generated first is treated
-                        # as a valid nan value.
-                        nan_idx = series.str.startswith(NUMERIC_NA_TOKEN)
-                        series.loc[nan_idx] = "NaN"
+                        nan_idx = None
+                        if self.column_has_missing[c]:
+                            # Values where nan was generated first is treated
+                            # as a valid nan value.
+                            nan_idx = series.str.startswith(NUMERIC_NA_TOKEN)
+                            series.loc[nan_idx] = "NaN"
 
                         # Removed indices of invalid values. Invalid values
                         # are those where nan was generated somewhere after
@@ -136,16 +144,16 @@ class REaLSampler:
                         # The initially changed "NaN" above will be
                         # included by the pattern. Make sure we add indices
                         # back to the valid_idx list.
-                        valid_idx = valid_idx.difference(
-                            np.where(
-                                series.str.contains(re_invalid_pattern, regex=True)
-                            )[0]
-                        ).union(np.where(nan_idx)[0])
+                        invalid_idx = series.str.contains(re_invalid_pattern, regex=True)
+                        valid_idx = valid_idx.difference(np.where(invalid_idx)[0])
+
+                        if nan_idx is not None:
+                            valid_idx = valid_idx.union(np.where(nan_idx)[0])
 
                         # Temporarily set them to NaN, but we will remove these
                         # observations later.
                         series.loc[
-                            series.str.contains(re_invalid_pattern, regex=True)
+                            invalid_idx
                         ] = "NaN"
                 elif pd.api.types.is_object_dtype(d):
                     # In case the values in this column is "integral"
@@ -162,12 +170,26 @@ class REaLSampler:
                     except ValueError:
                         pass
 
-                _synth_df.append(series.astype(d))
+                if series.dtype != d:
+                    # To speed up a bit, only cast if the dtypes
+                    # is not the same as the expected type.
+                    series = series.astype(d)
+
+                _synth_df.append(series)
             except Exception:  # nolint
                 _synth_df.append(synth_df[c])
 
         synth_df = pd.concat(_synth_df, axis=1)
-        synth_df = synth_df.replace("<NA>", None).replace("NaN", None)
+
+        # We expect columns that have no missing values in the
+        # training data should have no <NA> in them.
+        with_missing_cols = [col for col, missing in self.column_has_missing.items() if missing]
+        if with_missing_cols:
+            synth_df[with_missing_cols] = (
+                synth_df[with_missing_cols]
+                .replace("<NA>", None)
+                .replace("NaN", None)
+            )
 
         if len(valid_idx) > 0:
             synth_df = synth_df.iloc[sorted(valid_idx)]
@@ -225,6 +247,11 @@ class REaLSampler:
 
         # Validate that the generated value for a column is correct.
         # Use the column identifier as basis for the validation.
+        # Is this useful when we actually filter the vocabulary
+        # during generation???
+        # Let's try removing this for now and see what happens... XD
+        # We remove this because the operations here are expensive, and
+        # slow down the sampling.
         for col in synth_sample.columns:
             valid_mask.append(synth_sample[col].str.startswith(col))
 
@@ -387,7 +414,12 @@ class REaLSampler:
                 samples, columns=self.processed_columns, index=group_ids
             )
 
-        synth_sample = self._validate_synth_sample(synth_sample)
+        # # Is this useful when we actually filter the vocabulary
+        # # during generation???
+        # # Let's try removing this for now and see what happens... XD
+        # # We remove this because the operations here are expensive, and
+        # # slow down the sampling.
+        # synth_sample = self._validate_synth_sample(synth_sample)
 
         # Extract the values
         for col in synth_sample.columns:
@@ -440,6 +472,7 @@ class TabularSampler(REaLSampler):
         columns: List,
         datetime_columns: List,
         column_dtypes: Dict,
+        column_has_missing: Dict,
         drop_na_cols: List,
         col_transform_data: Dict,
         random_state: Optional[int] = 1029,
@@ -456,6 +489,7 @@ class TabularSampler(REaLSampler):
             columns,
             datetime_columns,
             column_dtypes,
+            column_has_missing,
             drop_na_cols,
             col_transform_data,
             random_state,
@@ -483,6 +517,7 @@ class TabularSampler(REaLSampler):
             columns=rtf_model.columns,
             datetime_columns=rtf_model.datetime_columns,
             column_dtypes=rtf_model.column_dtypes,
+            column_has_missing=rtf_model.column_has_missing,
             drop_na_cols=rtf_model.drop_na_cols,
             col_transform_data=rtf_model.col_transform_data,
             random_state=rtf_model.random_state,
@@ -588,6 +623,9 @@ class TabularSampler(REaLSampler):
                     **generate_kwargs,
                 )
 
+                self.total_gen_samples += len(sample_outputs)
+                self.invalid_gen_samples += len(sample_outputs)
+
                 try:
                     synth_sample = self._processes_sample(
                         sample_outputs=sample_outputs,
@@ -595,6 +633,8 @@ class TabularSampler(REaLSampler):
                         validator=validator,
                     )
                     empty_limit = continuous_empty_limit
+                    self.invalid_gen_samples -= len(synth_sample)
+
                 except SampleEmptyError as exc:
                     logging.warning("This batch returned an empty valid synth_sample!")
                     empty_limit -= 1
@@ -616,6 +656,8 @@ class TabularSampler(REaLSampler):
         )
         synth_df = synth_df.reset_index(drop="index")
 
+        print(f"Generated {self.invalid_gen_samples} invalid samples out of total {self.total_gen_samples} samples generated. Sampling efficiency is: {100 * (1 -  self.invalid_gen_samples / self.total_gen_samples):.4f}%")
+
         return synth_df
 
     def predict(
@@ -712,6 +754,7 @@ class RelationalSampler(REaLSampler):
         columns: List,
         datetime_columns: List,
         column_dtypes: Dict,
+        column_has_missing: Dict,
         drop_na_cols: List,
         col_transform_data: Dict,
         in_col_transform_data: Dict,
@@ -729,6 +772,7 @@ class RelationalSampler(REaLSampler):
             columns,
             datetime_columns,
             column_dtypes,
+            column_has_missing,
             drop_na_cols,
             col_transform_data,
             random_state,
@@ -758,6 +802,7 @@ class RelationalSampler(REaLSampler):
             columns=rtf_model.columns,
             datetime_columns=rtf_model.datetime_columns,
             column_dtypes=rtf_model.column_dtypes,
+            column_has_missing=rtf_model.column_has_missing,
             drop_na_cols=rtf_model.drop_na_cols,
             col_transform_data=rtf_model.col_transform_data,
             in_col_transform_data=rtf_model.in_col_transform_data,
